#!/usr/bin/perl

use strict;
use CGI::Carp qw(fatalsToBrowser);
use CGI;
use Encode qw(decode_utf8);

BEGIN { require "wakautils.pl" }

use constant MAPS_KEY => ""; # Get a Google Maps API key from http://www.google.com/apis/maps/signup.html and put it here.
use constant STATIC_FILE => ""; # Set this to a local filename if you want the script to only show one specific file.
use constant MAX_DOWNLOAD_SIZE => 65536; # Limit the maximum size the script will download over HTTP here.

use constant VIEW_TEMPLATE => compile_template(q{
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xmlns:v="urn:schemas-microsoft-com:vml">

<head>
<meta http-equiv="content-type" content="text/html; charset=UTF-8" />

<title>KML viewer</title>

<style type="text/css">
v\:* {
	behavior:url(#default#VML);
}

#map {
	position: absolute;
	left: 8px;
	top: 8px;
	border: 2px solid black;
	width:600px;
	height:600px;
}
body>#map {
	width: auto;
	height: auto;
	right: 264px;
	bottom: 8px;
}
#list {
	position: absolute;
	left: 616px;
	top: 8px;
	width: 230px;
	height: 584px;
	border: 2px solid #003f00;
	padding: 8px;
	background: #efffef;
	font-family: sans-serif;
	font-size: 10pt;
	overflow: auto;
}
body>#list {
	height: auto;
	left: auto;
	right: 8px;
	bottom: 8px;
}
#list h2 {
	font-size: 12pt;
	margin: 0px 0px 6px 0px;
	padding: 0px;
	color: #003f00;
	font-weight: bold;
}
#list a {
	display:block;
	color: #003f00;
	text-decoration: none;
	margin-bottom: 2px;
}
#list a:hover {
	color: #007f00;
}
.bubble {
	width: 300px;
	font-size: 10pt;
}
.bubble strong {
	font-family: sans-serif;
}
</style>

<script src="http://maps.google.com/maps?file=api&amp;v=2&amp;key=<const MAPS_KEY>" type="text/javascript">
</script>

<script type="text/javascript">
//<![CDATA[

function init()
{
	if(!GBrowserIsCompatible()) return;

	var map=new GMap(document.getElementById("map"));

	map.addControl(new GLargeMapControl());
	map.addControl(new GMapTypeControl());
	map.addControl(new GScaleControl());

	resize_map(map,<var $minlong>,<var $maxlong>,<var $minlat>,<var $maxlat>);

	var marker;
	<loop $places>
	map.addOverlay(make_marker(map,<var js_string($name)>,<var js_string($description)>,<var js_array(@{$coords})>));
	</loop>
}

function make_marker(map,n,desc,coords)
{
	var point=new GLatLng(coords[1],coords[0]);
	var marker=new GMarker(point);

	var html='<div class="bubble">';
	html+='<strong>'+n+'</strong>';
	if(desc) html+='<p>'+desc+'</p>';
	html+='</div>';

	GEvent.addListener(marker,'click',function() {
		marker.openInfoWindowHtml(html);
	});

	if(coords.length>2)
	{
		var latlngs=[];

		for(var i=0;i<coords.length/2;i++)
		latlngs.push(new GLatLng(coords[2*i+1],coords[2*i]));

		var polyline=new GPolyline(latlngs,"#ff7f7f",6);
		map.addOverlay(polyline);
	}

	var list=document.getElementById("list");
	var link=document.createElement("a");
	link.innerHTML=n;
	link.href="javascript:void(0)";
	link.onclick=function() {
		map.recenterOrPanToLatLng(point);
		marker.openInfoWindowHtml(html);
	}
	list.appendChild(link);

	return marker;
}

function resize_map(map,minlong,maxlong,minlat,maxlat)
{
	var center=new GPoint((maxlong+minlong)/2,(maxlat+minlat)/2);

	for(var i=0;i<16;i++)
	{
		map.centerAndZoom(center,i);
		var bounds=map.getBoundsLatLng();
		if(bounds.minX<minlong&&bounds.maxX>maxlong&&bounds.minY<minlat&&bounds.maxY>maxlat) break;
	}
}

//]]>
</script>

</head>
<body onload="init()" onunload="GUnload()">
<div id="map"></div>
<div id="list">
<h2>Locations</h2>
</div>
</body>
</html>

},1);

my $kml;

if(STATIC_FILE)
{
	$kml=read_file(STATIC_FILE);
}
else
{
	my $query=new CGI;
	my $url=$query->param("url");
	my $file=$query->param("file");
	my $text=$query->param("text");

	if($url) { $kml=get_http($url,MAX_DOWNLOAD_SIZE,"","kmlcache/cache-") }
	elsif($file) { $kml=read_file($file); }
	elsif($text) { $kml=$text }
}

$kml=z_unpack($kml) if $kml=~/^PK/;

my @places=parse_kml($kml);
my ($minlong,$maxlong,$minlat,$maxlat)=find_extents(@places);


print "Content-Type: text/html; charset=utf-8\n";
print "\n";
print VIEW_TEMPLATE->(
	minlong=>$minlong,
	maxlong=>$maxlong,
	minlat=>$minlat,
	maxlat=>$maxlat,
	places=>\@places,
);

#use XML::Twig;

sub parse_kml($)
{
	my ($kml)=@_;
	my @placemarks;

#	sub placemark {
#		my ($twig,$placemark)=@_;
#		my %mark;
#
#		my $name_el=$placemark->first_child("name");
#		my $description_el=$placemark->first_child("description");
#		my $style_el=$placemark->first_child("styleUrl");
#		#my $pointcoord_el=$placemark->first_child("Point/coordinates");
#		#my $linecoord_el=$placemark->first_child("LineString/coordinates");
#		my ($coord_el)=$placemark->descendants("coordinates");
#
#		$twig->purge;
#
#		return if $style_el and $style_el->text eq "#track";
#
#		$mark{name}=$name_el->text if $name_el;
#		$mark{description}=$description_el->text if $description_el;
#		$mark{coords}=[map m/(.*?),(.*?),/,grep $_,split / +/,$coord_el->text];
#
#		push @placemarks,\%mark;
#	}
#
#	sub start_folder {
#	}
#
#	sub folder_name {
#	}
#
#	my $t=XML::Twig->new(
#		twig_handlers=>{
#			"Placemark" => \&placemark,
#			"Folder/name" => \&folder_name,
#		},
#		twig_start_handlers=>{
#			"Folder" => \&start_folder,
#		},
#	);
#
#	$t->parsestring($kml);
#
#	return @placemarks;

	my $url_re=url_regexp();

	$kml=decode_utf8($kml);

	return map
	{
		my ($name)=m!<name>(.*?)</name>!s;
		my ($coords)=m!<coordinates>(.*?)</coordinates>!s;
		my ($description)=m!<description>(.*?)</description>!s;
		my ($style)=m!<styleUrl>(.*?)</styleUrl>!s;
		my %mark;

		$description=~s{$url_re}{<a href="$1" target="_blank">$1</a>$2}sg;

		if($style eq "#track") { () }
		else
		{
			$mark{name}=$name;
			$mark{description}=$description;
			$mark{coords}=[map m/(.*?),(.*?),/,grep $_,split / +/,$coords];

			\%mark;
		}
	} $kml=~m!<Placemark>(.*?)</Placemark>!gs;
}

sub find_extents(@)
{
	my ($minlong,$maxlong,$minlat,$maxlat)=(1000,-1000,1000,-1000);
	foreach my $place (@_)
	{
		my ($long,$lat)=@{$$place{coords}};
		$minlong=$long if $long<$minlong;
		$maxlong=$long if $long>$maxlong;
		$minlat=$lat if $lat<$minlat;
		$maxlat=$lat if $lat>$maxlat;
	}
	return ($minlong,$maxlong,$minlat,$maxlat);
}

sub read_file($)
{
	my $file=shift;
	my ($filehandle,$contents,$buffer);

	if(ref $file eq "Fh") { $filehandle=$file }
	else { open $filehandle,$file or return undef }

	binmode $filehandle;
	while(read $filehandle,$buffer,4096) { $contents.=$buffer }

	return $contents;
}

sub z_unpack($)
{
	my $file=shift;
	my $name="/tmp/kmlview".make_random_string(10);

	open ZCAT,"|zcat >$name" or die "Couldn't run zcat to decompress file";
	print ZCAT $file;
	close ZCAT;

	my $decompressed=read_file($name);

	unlink $file;

	return $decompressed;
}
